@@ -71,6 +71,9 @@ mutt_window_t *MuttMessageWindow = NULL;
 mutt_window_t *MuttSidebarWindow = NULL;
 #endif
 
+static void reflow_message_window_rows (int mw_rows);
+
+
 void mutt_refresh (void)
 {
   /* don't refresh when we are waiting for a child. */
@@ -94,7 +97,7 @@ void mutt_need_hard_redraw (void)
 {
   keypad (stdscr, TRUE);
   clearok (stdscr, TRUE);
-  set_option (OPTNEEDREDRAW);
+  mutt_set_current_menu_redraw_full ();
 }
 
 event_t mutt_getch (void)
@@ -162,6 +165,15 @@ int _mutt_get_field (const char *field, char *buf, size_t buflen, int complete,
   
   do
   {
+#if defined (USE_SLANG_CURSES) || defined (HAVE_RESIZETERM)
+    if (SigWinch)
+    {
+      SigWinch = 0;
+      mutt_resize_screen ();
+      clearok(stdscr, TRUE);
+      mutt_current_menu_redraw ();
+    }
+#endif
     mutt_window_clearline (MuttMessageWindow, 0);
     SETCOLOR (MT_COLOR_PROMPT);
     addstr ((char *)field); /* cast to get around bad prototypes */
@@ -173,7 +185,7 @@ int _mutt_get_field (const char *field, char *buf, size_t buflen, int complete,
   while (ret == 1);
   mutt_window_clearline (MuttMessageWindow, 0);
   mutt_free_enter_state (&es);
-  
+
   return (ret);
 }
 
@@ -220,8 +232,9 @@ int mutt_yesorno (const char *msg, int def)
   char *yes = _("yes");
   char *no = _("no");
   char *answer_string;
-  size_t answer_string_len;
-  size_t msglen;
+  int answer_string_wid, msg_wid;
+  size_t trunc_msg_len;
+  int redraw = 1, prompt_lines = 1;
 
 #ifdef HAVE_LANGINFO_YESEXPR
   char *expr;
@@ -239,8 +252,6 @@ int mutt_yesorno (const char *msg, int def)
             !REGCOMP (&reno, expr, REG_NOSUB);
 #endif
 
-  mutt_window_clearline (MuttMessageWindow, 0);
-
   /*
    * In order to prevent the default answer to the question to wrapped
    * around the screen in the even the question is wider than the screen,
@@ -248,19 +259,55 @@ int mutt_yesorno (const char *msg, int def)
    * to fit.
    */
   safe_asprintf (&answer_string, " ([%s]/%s): ", def == MUTT_YES ? yes : no, def == MUTT_YES ? no : yes);
-  answer_string_len = mutt_strwidth (answer_string);
-  /* maxlen here is sort of arbitrary, so pick a reasonable upper bound */
-  msglen = mutt_wstr_trunc (msg, 4*MuttMessageWindow->cols, MuttMessageWindow->cols - answer_string_len, NULL);
-  SETCOLOR (MT_COLOR_PROMPT);
-  addnstr (msg, msglen);
-  addstr (answer_string);
-  NORMAL_COLOR;
-  FREE (&answer_string);
+  answer_string_wid = mutt_strwidth (answer_string);
+  msg_wid = mutt_strwidth (msg);
 
   FOREVER
   {
+    if (redraw || SigWinch)
+    {
+      redraw = 0;
+#if defined (USE_SLANG_CURSES) || defined (HAVE_RESIZETERM)
+      if (SigWinch)
+      {
+        SigWinch = 0;
+        mutt_resize_screen ();
+        clearok (stdscr, TRUE);
+        mutt_current_menu_redraw ();
+      }
+#endif
+      if (MuttMessageWindow->cols)
+      {
+        prompt_lines = (msg_wid + answer_string_wid + MuttMessageWindow->cols - 1) /
+          MuttMessageWindow->cols;
+        prompt_lines = MAX (1, MIN (3, prompt_lines));
+      }
+      if (prompt_lines != MuttMessageWindow->rows)
+      {
+        reflow_message_window_rows (prompt_lines);
+        mutt_current_menu_redraw ();
+      }
+
+      /* maxlen here is sort of arbitrary, so pick a reasonable upper bound */
+      trunc_msg_len = mutt_wstr_trunc (msg, 4 * prompt_lines * MuttMessageWindow->cols,
+                                       prompt_lines * MuttMessageWindow->cols - answer_string_wid,
+                                       NULL);
+
+      mutt_window_move (MuttMessageWindow, 0, 0);
+      SETCOLOR (MT_COLOR_PROMPT);
+      addnstr (msg, trunc_msg_len);
+      addstr (answer_string);
+      NORMAL_COLOR;
+      mutt_window_clrtoeol (MuttMessageWindow);
+    }
+
     mutt_refresh ();
+    /* SigWinch is not processed unless timeout is set */
+    timeout (30 * 1000);
     ch = mutt_getch ();
+    timeout (-1);
+    if (ch.ch == -2)
+      continue;
     if (CI_is_return (ch.ch))
       break;
     if (ch.ch < 0)
@@ -297,6 +344,8 @@ int mutt_yesorno (const char *msg, int def)
     }
   }
 
+  FREE (&answer_string);
+
 #ifdef HAVE_LANGINFO_YESEXPR    
   if (reyes_ok)
     regfree (& reyes);
@@ -304,6 +353,14 @@ int mutt_yesorno (const char *msg, int def)
     regfree (& reno);
 #endif
 
+  if (MuttMessageWindow->rows != 1)
+  {
+    reflow_message_window_rows (1);
+    mutt_current_menu_redraw ();
+  }
+  else
+    mutt_window_clearline (MuttMessageWindow, 0);
+
   if (def != -1)
   {
     addstr ((char *) (def == MUTT_YES ? yes : no));
@@ -548,6 +605,33 @@ void mutt_reflow_windows (void)
     MuttIndexWindow->col_offset += SidebarWidth;
   }
 #endif
+
+  mutt_set_current_menu_redraw_full ();
+  /* the pager menu needs this flag set to recalc lineInfo */
+  mutt_set_current_menu_redraw (REDRAW_FLOW);
+}
+
+static void reflow_message_window_rows (int mw_rows)
+{
+  MuttMessageWindow->rows = mw_rows;
+  MuttMessageWindow->row_offset = LINES - mw_rows;
+
+  MuttStatusWindow->row_offset = option (OPTSTATUSONTOP) ? 0 : LINES - mw_rows - 1;
+
+  if (option (OPTHELP))
+    MuttHelpWindow->row_offset = option (OPTSTATUSONTOP) ? LINES - mw_rows - 1 : 0;
+
+  MuttIndexWindow->rows = MAX(LINES - MuttStatusWindow->rows -
+			      MuttHelpWindow->rows - MuttMessageWindow->rows, 0);
+
+#ifdef USE_SIDEBAR
+  if (option (OPTSIDEBAR))
+    MuttSidebarWindow->rows = MuttIndexWindow->rows;
+#endif
+
+  /* We don't also set REDRAW_FLOW because this function only
+   * changes rows and is a temporary adjustment. */
+  mutt_set_current_menu_redraw_full ();
 }
 
 int mutt_window_move (mutt_window_t *win, int row, int col)
@@ -733,7 +817,7 @@ int mutt_do_pager (const char *banner,
   return rc;
 }
 
-int _mutt_enter_fname (const char *prompt, char *buf, size_t blen, int *redraw, int buffy, int multiple, char ***files, int *numfiles)
+int _mutt_enter_fname (const char *prompt, char *buf, size_t blen, int buffy, int multiple, char ***files, int *numfiles)
 {
   event_t ch;
 
@@ -758,7 +842,6 @@ int _mutt_enter_fname (const char *prompt, char *buf, size_t blen, int *redraw,
     buf[0] = 0;
     _mutt_select_file (buf, blen, MUTT_SEL_FOLDER | (multiple ? MUTT_SEL_MULTI : 0), 
 		       files, numfiles);
-    *redraw = REDRAW_FULL;
   }
   else
   {
@@ -769,7 +852,6 @@ int _mutt_enter_fname (const char *prompt, char *buf, size_t blen, int *redraw,
     if (_mutt_get_field (pc, buf, blen, (buffy ? MUTT_EFILE : MUTT_FILE) | MUTT_CLEAR, multiple, files, numfiles)
 	!= 0)
       buf[0] = 0;
-    MAYBE_REDRAW (*redraw);
     FREE (&pc);
   }
 
@@ -871,16 +953,48 @@ int mutt_multi_choice (char *prompt, char *letters)
 {
   event_t ch;
   int choice;
+  int redraw = 1, prompt_lines = 1;
   char *p;
 
-  SETCOLOR (MT_COLOR_PROMPT);
-  mutt_window_mvaddstr (MuttMessageWindow, 0, 0, prompt);
-  NORMAL_COLOR;
-  mutt_window_clrtoeol (MuttMessageWindow);
   FOREVER
   {
+    if (redraw || SigWinch)
+    {
+      redraw = 0;
+#if defined (USE_SLANG_CURSES) || defined (HAVE_RESIZETERM)
+      if (SigWinch)
+      {
+        SigWinch = 0;
+        mutt_resize_screen ();
+        clearok (stdscr, TRUE);
+        mutt_current_menu_redraw ();
+      }
+#endif
+      if (MuttMessageWindow->cols)
+      {
+        prompt_lines = (mutt_strwidth (prompt) + MuttMessageWindow->cols - 1) /
+          MuttMessageWindow->cols;
+        prompt_lines = MAX (1, MIN (3, prompt_lines));
+      }
+      if (prompt_lines != MuttMessageWindow->rows)
+      {
+        reflow_message_window_rows (prompt_lines);
+        mutt_current_menu_redraw ();
+      }
+
+      SETCOLOR (MT_COLOR_PROMPT);
+      mutt_window_mvaddstr (MuttMessageWindow, 0, 0, prompt);
+      NORMAL_COLOR;
+      mutt_window_clrtoeol (MuttMessageWindow);
+    }
+
     mutt_refresh ();
+    /* SigWinch is not processed unless timeout is set */
+    timeout (30 * 1000);
     ch  = mutt_getch ();
+    timeout (-1);
+    if (ch.ch == -2)
+      continue;
     /* (ch.ch == 0) is technically possible.  Treat the same as < 0 (abort) */
     if (ch.ch <= 0 || CI_is_return (ch.ch))
     {
@@ -904,7 +1018,13 @@ int mutt_multi_choice (char *prompt, char *letters)
     }
     BEEP ();
   }
-  mutt_window_clearline (MuttMessageWindow, 0);
+  if (MuttMessageWindow->rows != 1)
+  {
+    reflow_message_window_rows (1);
+    mutt_current_menu_redraw ();
+  }
+  else
+    mutt_window_clearline (MuttMessageWindow, 0);
   mutt_refresh ();
   return choice;
 }
