@@ -56,7 +56,7 @@
   if ((CurrentMenu == MENU_PAGER) && (idx >= 0) &&	\
 	    (MuttVars[idx].flags & R_RESORT)) \
 	{ \
-	  snprintf (err->data, err->dsize, \
+	  snprintf (err->data, err->dsize, "%s", \
 	    _("Not available in this menu.")); \
 	  return (-1); \
 	}
@@ -524,7 +524,7 @@ static int add_to_replace_list (REPLACE_LIST **list, const char *pat, const char
 
   if (t->nmatch > t->rx->rx->re_nsub)
   {
-    snprintf (err->data, err->dsize, _("Not enough subexpressions for "
+    snprintf (err->data, err->dsize, "%s", _("Not enough subexpressions for "
                                        "template"));
     remove_from_replace_list(list, pat);
     return -1;
@@ -1298,8 +1298,6 @@ static int parse_attachments (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER
     print_attach_list(AttachExclude, '-', "A");
     print_attach_list(InlineAllow,   '+', "I");
     print_attach_list(InlineExclude, '-', "I");
-    set_option (OPTFORCEREDRAWINDEX);
-    set_option (OPTFORCEREDRAWPAGER);
     mutt_any_key_to_continue (NULL);
     return 0;
   }
@@ -1443,7 +1441,7 @@ static int parse_unalias (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *er
       {
 	for (tmp = Aliases; tmp ; tmp = tmp->next) 
 	  tmp->del = 1;
-	set_option (OPTFORCEREDRAWINDEX);
+	mutt_set_current_menu_redraw_full ();
       }
       else
 	mutt_free_alias (&Aliases);
@@ -1457,7 +1455,7 @@ static int parse_unalias (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *er
 	  if (CurrentMenu == MENU_ALIAS)
 	  {
 	    tmp->del = 1;
-	    set_option (OPTFORCEREDRAWINDEX);
+	    mutt_set_current_menu_redraw_full ();
 	    break;
 	  }
 
@@ -1518,7 +1516,7 @@ static int parse_alias (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
     /* override the previous value */
     rfc822_free_address (&tmp->addr);
     if (CurrentMenu == MENU_ALIAS)
-      set_option (OPTFORCEREDRAWINDEX);
+      mutt_set_current_menu_redraw_full ();
   }
 
   mutt_extract_token (buf, s, MUTT_TOKEN_QUOTE | MUTT_TOKEN_SPACE | MUTT_TOKEN_SEMICOLON);
@@ -1796,9 +1794,14 @@ static void mutt_restore_default (struct option_t *p)
   }
 
   if (p->flags & R_INDEX)
-    set_option (OPTFORCEREDRAWINDEX);
+    mutt_set_menu_redraw_full (MENU_MAIN);
   if (p->flags & R_PAGER)
-    set_option (OPTFORCEREDRAWPAGER);
+    mutt_set_menu_redraw_full (MENU_PAGER);
+  if (p->flags & R_PAGER_FLOW)
+  {
+    mutt_set_menu_redraw_full (MENU_PAGER);
+    mutt_set_menu_redraw (MENU_PAGER, REDRAW_FLOW);
+  }
   if (p->flags & R_RESORT_SUB)
     set_option (OPTSORTSUBTHREADS);
   if (p->flags & R_RESORT)
@@ -1811,8 +1814,10 @@ static void mutt_restore_default (struct option_t *p)
     mutt_reflow_windows ();
 #ifdef USE_SIDEBAR
   if (p->flags & R_SIDEBAR)
-    SidebarNeedsRedraw = 1;
+    mutt_set_current_menu_redraw (REDRAW_SIDEBAR);
 #endif
+  if (p->flags & R_MENU)
+    mutt_set_current_menu_redraw_full ();
 }
 
 static size_t escape_string (char *dst, size_t len, const char* src)
@@ -1983,8 +1988,6 @@ static int parse_setenv(BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
 
     if (found)
     {
-      set_option (OPTFORCEREDRAWINDEX);
-      set_option (OPTFORCEREDRAWPAGER);
       mutt_any_key_to_continue (NULL);
       return 0;
     }
@@ -2094,13 +2097,13 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
     {
       if (query || unset || inv)
       {
-	snprintf (err->data, err->dsize, _("prefix is illegal with reset"));
+	snprintf (err->data, err->dsize, "%s", _("prefix is illegal with reset"));
 	return (-1);
       }
 
       if (s && *s->dptr == '=')
       {
-	snprintf (err->data, err->dsize, _("value is illegal with reset"));
+	snprintf (err->data, err->dsize, "%s", _("value is illegal with reset"));
 	return (-1);
       }
      
@@ -2108,13 +2111,12 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
       {
 	if (CurrentMenu == MENU_PAGER)
 	{
-	  snprintf (err->data, err->dsize, _("Not available in this menu."));
+	  snprintf (err->data, err->dsize, "%s", _("Not available in this menu."));
 	  return (-1);
 	}
 	for (idx = 0; MuttVars[idx].option; idx++)
 	  mutt_restore_default (&MuttVars[idx]);
-	set_option (OPTFORCEREDRAWINDEX);
-	set_option (OPTFORCEREDRAWPAGER);
+	mutt_set_current_menu_redraw_full ();
 	set_option (OPTSORTSUBTHREADS);
 	set_option (OPTNEEDRESORT);
 	set_option (OPTRESORTINIT);
@@ -2136,7 +2138,7 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
       {
 	if (unset || inv || query)
 	{
-	  snprintf (err->data, err->dsize, _("Usage: set variable=yes|no"));
+	  snprintf (err->data, err->dsize, "%s", _("Usage: set variable=yes|no"));
 	  return (-1);
 	}
 
@@ -2148,7 +2150,7 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
 	  unset = 1;
 	else
 	{
-	  snprintf (err->data, err->dsize, _("Usage: set variable=yes|no"));
+	  snprintf (err->data, err->dsize, "%s", _("Usage: set variable=yes|no"));
 	  return (-1);
 	}
       }
@@ -2580,9 +2582,14 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
     if (!myvar)
     {
       if (MuttVars[idx].flags & R_INDEX)
-        set_option (OPTFORCEREDRAWINDEX);
+        mutt_set_menu_redraw_full (MENU_MAIN);
       if (MuttVars[idx].flags & R_PAGER)
-        set_option (OPTFORCEREDRAWPAGER);
+        mutt_set_menu_redraw_full (MENU_PAGER);
+      if (MuttVars[idx].flags & R_PAGER_FLOW)
+      {
+        mutt_set_menu_redraw_full (MENU_PAGER);
+        mutt_set_menu_redraw (MENU_PAGER, REDRAW_FLOW);
+      }
       if (MuttVars[idx].flags & R_RESORT_SUB)
         set_option (OPTSORTSUBTHREADS);
       if (MuttVars[idx].flags & R_RESORT)
@@ -2595,8 +2602,10 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
         mutt_reflow_windows ();
 #ifdef USE_SIDEBAR
       if (MuttVars[idx].flags & R_SIDEBAR)
-        SidebarNeedsRedraw = 1;
+        mutt_set_current_menu_redraw (REDRAW_SIDEBAR);
 #endif
+      if (MuttVars[idx].flags & R_MENU)
+        mutt_set_current_menu_redraw_full ();
     }
   }
   return (r);
